@@ -6,21 +6,78 @@ import logging
 import traceback
 import typing
 import uuid
-from langchain_core.messages import AIMessage, HumanMessage, ToolMessage, ToolCall
+from langchain_core.messages import AIMessage, HumanMessage, ToolMessage
 
 from .agent import Agent, BaseMessage, SystemMessage, AgentResponse
-from .chat_history import ChatHistory, SummaryRecord, MessageRecord
+from .chat_history import SummaryRecord, MessageRecord
 from .exceptions import ContextWindowExceededError
 from .prompt import build_prompt, build_all_tool_names
 from .tools import ask_user
-from .tool_utils import make_tool_dict, sanitize_toolname, tool, AgentRef
+from .tool_utils import make_tool_dict, tool, AgentRef
 from .models.base import BaseArchytasModel
-from .utils import extract_json, ensure_async
+from .utils import ensure_async
 
 
 logger = logging.Logger("archytas")
 
 
+def format_tool_result_for_display(result: str | list[dict] | dict) -> str:
+    """
+    Format tool result for display in terminal/logs.
+
+    Converts multimodal LangChain format to readable string.
+
+    Args:
+        result: Tool execution result (string or LangChain multimodal format)
+
+    Returns:
+        Formatted string for display
+    """
+    match result:
+        case str():
+            return result
+
+        case dict() if "type" in result:
+            # Single content block
+            return format_tool_result_for_display([result])
+
+        case list():
+            parts = []
+            for item in result:
+                if not isinstance(item, dict):
+                    continue
+
+                match item.get("type"):
+                    case "text":
+                        parts.append(item.get("text", ""))
+
+                    case "image_url":
+                        image_url = item.get("image_url", {})
+                        if isinstance(image_url, dict):
+                            url = image_url.get("url", "")
+                        else:
+                            url = str(image_url)
+
+                        # Show preview for base64, full URL for remote
+                        if url.startswith("data:"):
+                            # Extract mime type and show preview
+                            if ";" in url:
+                                mime_type = url.split(";")[0].replace("data:", "")
+                                parts.append(f"[Image: {mime_type}, base64 data]")
+                            else:
+                                parts.append("[Image: base64 data]")
+                        else:
+                            parts.append(f"[Image: {url}]")
+
+                    case _:
+                        parts.append(f"[{item.get('type', 'unknown')}]")
+
+            return "\n".join(parts) if parts else ""
+
+        case _:
+            return str(result)
+
+
 class Undefined:
     ...
 
@@ -59,9 +116,10 @@ Extra details:
 
 
 def catch_failure(fn):
-    def handle_error(messages: list[BaseMessage], error: Exception, agent: "ReActAgent|None" = None):
+    def handle_error(message_records: list[MessageRecord], error: Exception, agent: "ReActAgent|None" = None):
         last_ai_message = None
         seen_tool_message_ids = set()
+        messages = [record.message for record in message_records]
         for message in messages[::-1]:
             match message:
                 case AIMessage():
@@ -78,7 +136,6 @@ def catch_failure(fn):
                     content=str(error),
                     tool_call_id=missing_tool_id,
                 )
-                messages.append(error_message)
                 if agent:
                     agent.chat_history.add_message(error_message)
 
@@ -371,8 +428,33 @@ class ReActAgent(Agent):
                             }
                         )
 
+                        # Handle multimodal tool outputs (e.g., images from MCP tools)
+                        # OpenAI doesn't support images in tool messages, so we need to extract
+                        # images and send them in a separate user message
+                        tool_content = tool_output
+                        image_content = None
+
+                        if isinstance(tool_output, list):
+                            # Check if there are any images in the output
+                            text_blocks = []
+                            image_blocks = []
+
+                            for block in tool_output:
+                                if isinstance(block, dict):
+                                    if block.get("type") == "image_url":
+                                        image_blocks.append(block)
+                                    else:
+                                        text_blocks.append(block)
+
+                            # If there are images, separate them
+                            if image_blocks:
+                                # Tool message gets only text content
+                                tool_content = text_blocks if text_blocks else "Image generated"
+                                # Images will be sent in a follow-up user message
+                                image_content = image_blocks
+
                         tool_message = ToolMessage(
-                            content=tool_output,
+                            content=tool_content,
                             tool_call_id=tool_id,
                             artifact={
                                 "tool_name": tool_name,
@@ -383,6 +465,17 @@ class ReActAgent(Agent):
                         # Always add tool response before handling state changes to ensure message history is correct.
                         self.chat_history.add_message(tool_message)
 
+                        # If the tool returned images, inject them as a user message
+                        # This works around OpenAI's limitation that images can only be in user messages
+                        if image_content:
+                            image_message = HumanMessage(
+                                content=[
+                                    {"type": "text", "text": f"[Tool '{tool_name}' returned the following image(s):]"},
+                                    *image_content
+                                ]
+                            )
+                            self.chat_history.add_message(image_message)
+
                         # Have the agent observe the result, and get the next action
                         if self.verbose:
                             self.display_observation(tool_output)
@@ -523,7 +616,7 @@ class ReActAgent(Agent):
         )
 
         if not summary:
-            return f"Error: Unable to locate summary {summarized_record_uuid}"
+            return f"Error: Unable to locate summary {summary_record_uuid}"
 
         # This method of collecting records should ensure correct ordering
         summarized_records: list[MessageRecord[BaseMessage]] = [
@@ -565,11 +658,11 @@ Available summaries are:
             if isinstance(record.message, AIMessage) and record.message.tool_calls:
                 for tool_call in record.message.tool_calls:
                     response_parts.extend([
-                        f"""\
+                        """\
 --- Tool Call     ---
 """,
                         json.dumps(tool_call),
-                        f"""\
+                        """\
 --- End Tool Call ---
 """,
                     ])
